;; Problem 27
;; 27 September 2002

;; Euler published the remarkable quadratic formula:

;;   + n + 41

;; It turns out that the formula will produce 40 primes for the consecutive values n = 0 to 39. However, when n = 40, 402 + 40 + 41 = 40(40 + 1) + 41 is divisible by 41, and certainly when n = 41,   + 41 + 41 is clearly divisible by 41.

;; Using computers, the incredible formula    79n + 1601 was discovered, which produces 80 primes for the consecutive values n = 0 to 79. The product of the coefficients, 79 and 1601, is 126479.

;; Considering quadratics of the form:

;;       + an + b, where |a| 1000 and |b| 1000

;;     where |n| is the modulus/absolute value of n
;;     e.g. |11| = 11 and |4| = 4

;; Find the product of the coefficients, a and b, for the quadratic expression that produces the maximum number of primes for consecutive values of n, starting with n = 0.nn41n

(defun prime? (num)
  (declare (optimize speed)
	   (type fixnum num))
  (when (< num 0)
    (return-from prime? nil))
  (when (= 2 num)
    (return-from prime? t))
  (loop for fact = (ceiling (sqrt num))
	then (if (oddp fact)
		 (- fact 2)
		 (- fact 1))
	while (> fact 1)
	do
	(when (= 0 (mod num fact))
	  (return-from prime? nil)))
  (not (= 0 (mod num 2))))

(defun compute-value (nn aa bb)
  (+ (* nn nn)
     (* aa nn)
     bb))

(defun count-primes-produced (aa bb)
  (declare (optimize speed)
	   (type fixnum aa)
	   (type fixnum bb))
  (let ((cnt 0)
	(nn 0))
    (declare (type fixnum cnt)
	     (type fixnum nn))
    (loop do
	  (when (not (prime? (+ (* nn nn)
				(* aa nn)
				bb)))
	    (return nn))
	  (incf cnt)
	  (incf nn))))

(assert (equal 40 (count-primes-produced 1 41)))
(assert (equal 80 (count-primes-produced -79 1601)))

(defun valid-bs ()
  (let ((nums (list)))
    (loop for xx from -999 to 999
	  do
	  (when (prime? (abs xx))
	    (push xx nums)))
    (reverse nums)))

; (length (valid-bs))

(defvar *valid-bs* (valid-bs))

(defun find-best (start end)
  (let ((best-aa nil)
	(best-bb nil)
	(best-count 0))
    (loop for aa from start to end
	  do
	  (format t "~a~&" aa)
	  (loop for bb in *valid-bs*
		do
		(let ((cnt (count-primes-produced aa bb)))
		  (when (> cnt best-count)
		    (setf best-aa aa
			  best-bb bb
			  best-count cnt)
		    (when (> cnt 0)
		      (format t "aa:~a bb:~a cnt:~a~&"
			      aa bb cnt))
		    (format t "new best aa:~a bb:~a cnt:~a~&"
			    aa
			    bb
			    cnt)))))
    (format t "best: aa:~a bb:~a cnt:~a~&"
	    best-aa best-bb best-count)))

; (sb-thread:list-all-threads)
; (sb-thread:interrupt-thread)

; (find-best -1000 1000)
;; => best: aa:-61 bb:971 cnt:71
;; (* -61 971) -59231

; (find-best 0 2)

;; new best aa:-999 bb:61 cnt:1011
;; new best aa:-999 bb:61 cnt:1011

; (compute-value 1 -999 61)

; (prime? -937)

; (prime? 61)